/*
 * Copyright (c) 2008  Yuri Chernyavsky
 * Copyright (c) 2007  Gregor Richards
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef SMV_GUI_H
#define SMV_GUI_H

class Frontend;
class Parameter;

class Backend : public Gtk::Window {
 public:
  Backend(Frontend* frontend);
  virtual ~Backend();

  // Launch the encoder
  void launch(const char* inp, const char* out, const std::vector<Parameter*>& params);

  // Reap child and see return status
  int finish();

  // Gently interrupt the encoder
  // handle interrupt request from the Frontend
  void cancel();

  // Brutally kill the encoder
  // in case the encoder behaves oddly
  void stop();

  // Get message from the encoder.
  // This method is to be called by the Frontend
  // upon reciept of a message_dispatcher event
  msg_t get_message();

 private:
  Frontend* frontend_;
  void thread_function();
  std::queue<msg_t> message_queue_;
  // the message queue is protected by a mutex
  Glib::Mutex message_queue_mutex_;
  int child_pid;
  int child_fd;

 protected:
  void push_message(msg_t msg);
  void child_setup();
};

class Frontend : public Gtk::Window {  
 public:
  Frontend();
  Frontend(BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade::Xml>& glade_xml);
  virtual ~Frontend();

  // Proceed message
  void proceed_message(msg_t msg);

  // Message arrived event
  Glib::Dispatcher& message_dispatcher() {
    return message_dis_;
  }

  // Encoder started
  Glib::Dispatcher& start_dispatcher() {
    return start_dis_;
  }
  
  // Encoder finished
  Glib::Dispatcher& end_dispatcher() {
    return end_dis_;
  }

 private:
  Glib::Dispatcher message_dis_, start_dis_, end_dis_;
  Backend*  backend_;

  // This flag is for backend sanity check.
  // Note that the encoder may be a wrong executable
  // or not exist at all - we cannot know because
  // we run encoder within a shell.
  // So the encoder sends a special "hello" message
  // that lets us know we communicate with the proper
  // backend.
  // backend_status_ == 0  -> waiting for "hello" message
  // backend_status_ == 1  -> ok
  // backend_status_ == -1 -> sanity check failed
  int backend_status_;
  
  // This button can be either convert button
  // or cancel button, so we need a handle to its
  // connection object
  sigc::connection convert_button_connection_;
  
  // Encoder parameters - can be easily modified when
  // the encoder changes
  std::vector<Parameter*> parameters_;

  // The widgets - defined in the glade file
  Gtk::Entry *input_file_name_, *output_file_name_;
  Gtk::Button *input_file_chooser_button_, *output_file_chooser_button_;
  Gtk::FileChooserDialog *input_file_dialog_, *output_file_dialog_;
  Gtk::Button *save_param_button_, *default_param_button_;
  Gtk::ProgressBar *progress_bar_;
  Gtk::Button *convert_button_, *quit_button_;
  Gtk::Table *params_table_;
  Gtk::Image *convert_img_, *cancel_img_;

 protected:
  // Message strings
  const char *phasestr(phase_t p);
  const char *errstr(err_t e);
  bool on_delete_event(GdkEventAny* event);
  // Signal handlers
  void on_message_sig();
  void on_start_sig();
  void on_end_sig();
  // Button handlers
  void set_convert_button();
  void set_cancel_button();
  void on_convert();
  void on_quit();
  void on_cancel();
  void on_browse_input_file();
  void on_browse_output_file();
  // Init functions
  void init_file_dialogs();
  void init_parameters();
  void update_gui_after_init();
  void update_gui_upon_start_encoding();
  void update_gui_upon_end_encoding(int encoder_ok, int success);
  void update_output_filename(const std::string& input_filename);
  // Save encoder parameters
  void save_parameters();
  void set_default_parameters();
  void load_parameters();
};

// A generic encoder parameter interface
// The design is quite flexible and can be changed
// when encoder parameters change - see Frontend::init_parameters()
class Parameter {
 public:
  // Get displayed value
  virtual std::string get_val() = 0;
  // Set displayed value
  virtual void set_val(const std::string& val) = 0;
  // Get parameter name (as displayed in the gui)
  virtual std::string get_name() = 0;
  // Get flag name (as passed to the encoder (for instance -g)
  virtual std::string get_flag() = 0;
  // Get value as passed to the gui (for instance 96x128 for geometry)
  // The displayed value and the passed value can differ
  virtual std::string get_flag_and_val() = 0;
  // Set the default displayed value
  virtual void reset() = 0;
  // Get the widget
  virtual Gtk::Widget* get_widget() = 0;
  // Separate title from the widget
  virtual bool separate_title() = 0;
};

// Editable selection box
class EditableTextParameter : public Parameter {
 public:
  EditableTextParameter(const std::string& name, const std::string& flag,
			const std::vector<std::string>& all_vals,
			const std::string& initial_value);

  virtual ~EditableTextParameter() {}

  std::string get_val();
  std::string get_flag_and_val();
  void set_val(const std::string& val);
  std::string get_name() {
    return name_;
  }
  std::string get_flag() {
    return flag_;
  }
  Gtk::Widget* get_widget() {
    return widget_;
  }
  void reset();
  bool separate_title() {
    return true;
  }

 private:
  std::string name_, flag_, initial_value_;
  std::vector<std::string> all_vals_;
  Gtk::Widget *widget_;
};

// Non-editable selection box
class TextParameter : public Parameter {
 public:
  TextParameter(const std::string& name, 
		   const std::string& flag,
	 const std::vector<std::pair<std::string, std::string> >& sorted_vals,
	 const std::string& initial_value);

  virtual ~TextParameter() {}

  std::string get_val();
  std::string get_flag_and_val();
  void set_val(const std::string& val);
  std::string get_name() {
    return name_;
  }
  std::string get_flag() {
    return flag_;
  }
  Gtk::Widget* get_widget() {
    return widget_;
  }
  void reset();
  bool separate_title() {
    return true;
  }

 private:
  std::string name_, flag_, initial_value_;
  std::map<std::string, std::string> all_vals_;
  Gtk::Widget *widget_;
};

// SpinBox
class NumberParameter : public Parameter {
 public:
  NumberParameter(const std::string& name, const std::string& flag,
	    double min, double max, double step, double initial_value);

  virtual ~NumberParameter() {}

  std::string get_val();
  std::string get_flag_and_val();
  void set_val(const std::string& val);
  std::string get_name() {
    return name_;
  }
  std::string get_flag() {
    return flag_;
  }
  Gtk::Widget* get_widget() {
    return widget_;
  }
  void reset();
  bool separate_title() {
    return true;
  }

 private:
  std::string name_, flag_;
  double min_, max_, step_, initial_value_;
  Gtk::Widget *widget_;
};

#endif
